//TODO: "deep" epsilon closure + get_state

#include "Automaton.h"

#include <set>

Automaton::Automaton(set<char> sigma, Regexp* r)
{
  alphabet = sigma;

  if(r->is_epsilon)
    {
      state.push_back(new State());
      state.push_back(new State()); //Create 2 states
      succ[state[0]][new Label()] = state[1];
      //transition state0 --epsilon--> state1

      initial_state = state[0];
      final_state.push_back(state[1]); //no transition
    }

  if(r->is_letter)
    {
      state.push_back(new State());
      state.push_back(new State()); //Create 2 states
      succ[state[0]][new Label(r->letter)] = state[1];
      //transition state0 --letter--> state1

      initial_state = state[0];
      final_state.push_back(state[1]); 
    }

  if(r->is_star)
    {
      state.push_back(new State()); //initial state
      state.push_back(new State()); //final state

      Automaton* a = new Automaton(alphabet,r->reg1);
      for(int i=0; i<a->state.size(); i++)
	//For each state of a...
	{
	  state.push_back(a->state[i]); //copy state
	  succ[a->state[i]] = a->succ[a->state[i]]; //copy outgoing transitions
	}
      //Add the automaton for "reg"

      succ[state[0]][new Label()] = a->initial_state; //initial -> old_initial
      succ[a->initial_state][new Label()] = state[1]; //old_initial -> final
      succ[a->final_state[0]][new Label()] = a->initial_state; 
      //old_final -> old_initial

      initial_state = state[0];
      final_state.push_back(state[1]); //set initial and final states
    }

  if(r->is_concat)
    {
      Automaton* a1 = new Automaton(alphabet,r->reg1);
      Automaton* a2 = new Automaton(alphabet,r->reg2);

      for(int i=0; i<a1->state.size(); i++)
	//For each state of a1...
	{
	  state.push_back(a1->state[i]); //copy state
	  succ[a1->state[i]] = a1->succ[a1->state[i]]; //copy transitions
	}
      //Add the automaton for "reg1"

      for(int i=0; i<a2->state.size(); i++)
	//For each state of a2...
	{
	  state.push_back(a2->state[i]); //copy state
	  succ[a2->state[i]] = a2->succ[a2->state[i]]; //copy transitions
	}
      //Add the automaton for "reg2"

      succ[a1->final_state[0]][new Label()] = a2->initial_state;

      initial_state = a1->initial_state;
      final_state.push_back(a2->final_state[0]); 
    }

  if(r->is_union)
    {
      /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
      /* A COMPLETER                                                   */


      /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    }
}

int 
Automaton::state_rank(State* s)
{
  for(int i=0; i<state.size(); i++)
    if(state[i] == s)
      return i;
  cerr << "Automaton::state_rank: state not found" << endl;
  exit(1);
}


int
Automaton::is_final(State* s)
{
  int result;
  for(int i=0; i<final_state.size(); i++)
    if(final_state[i] == s)
      return true;
  return false;
}


set<State*> operator +(set<State*> s1, set<State*> s2)
{
  set<State*> result = s1;
  for(set<State*>::const_iterator it = s2.begin(); 
      it != s2.end(); ++it)
    result.insert(*it);

  return result;
}


set<State*> 
Automaton::epsilon_closure_(State* s, set<State*> reached)
{
  set<State*> result;
  result.insert(s);

  if(reached.find(s) != reached.end())
    return result;

  reached.insert(s);

  for(map<Label*, State*>::const_iterator it = succ[s].begin(); 
      it != succ[s].end(); ++it)
    if(it->first->is_epsilon)
      //For each epsilon-transition starting from s...
      result = result + epsilon_closure_(it->second,reached);;

  return result;
}


set<State*> 
Automaton::epsilon_closure(State* s)
{
  set<State*> reached;
  return epsilon_closure_(s,reached);
}


set<State*> 
Automaton::epsilon_closure(set<State*> state_set)
{
  set<State*> result;
  for(set<State*>::const_iterator it = state_set.begin(); 
      it != state_set.end(); ++it)
    //For each state of the set...
    result = result + epsilon_closure(*it);
  return result;
}

ostream& operator<< (ostream& sout, set<State*> s)
{
  sout << "{";
  for(set<State*>::const_iterator it = s.begin(); 
      it != s.end(); ++it)
    sout << *it << " ";
  sout << "}"; 
}

State* get_state(set<State*> actual_det_state, 
		 map<State*, set<State*> > actual_det_states)
{
  for(map<State*, set<State*> >::const_iterator it = actual_det_states.begin(); 
      it != actual_det_states.end(); ++it)
    //For each det_state -> vector<non_det_states> ..
    if(it->second == actual_det_state)
      return it->first;

  return 0; //not found => return NULL pointer (0)
}


Automaton*
Automaton::determinize()
{
  Automaton* result = new Automaton(alphabet);

  map<State*,set<State*> > non_det_states;

  result->state.push_back(new State());
  result->initial_state = result->state[0];
  non_det_states[result->state[0]] = epsilon_closure(initial_state);
  //initial state = closure(initial)

  vector<State*> state_stack;
  state_stack.push_back(result->state[0]);
  //start with initial state

  //
  // Build the automaton by depth-first traversal
  //
  while(state_stack.size() != 0)
    //For each (set of) state...
    {
      State* current_state = state_stack.back();
      state_stack.pop_back(); //pop a state

      //
      // Collect outgoing letters and corresponding target states 
      //
      set<char> outgoing_letters;
      map<char,set<State*> > target; //letter -> set of target states

      set<State*> actual_states = non_det_states[current_state];
      for(set<State*>::const_iterator it = actual_states.begin(); 
	  it != actual_states.end(); ++it)
	//For each (non-det) state in the current state...
	{
	  State* non_det_state = *it;
	  for(map<Label*,State*>::const_iterator it2 = 
		succ[non_det_state].begin(); 
	      it2 != succ[non_det_state].end(); ++it2)
	    //For each transition starting from non-det state... 
	    if(it2->first->is_letter)
	      {
		outgoing_letters.insert(it2->first->letter);
		target[it2->first->letter].insert(it2->second);
	      }
	}

      //
      // Then, update the target automaton
      //
      for(set<char>::const_iterator it = outgoing_letters.begin(); 
	  it != outgoing_letters.end(); ++it)
	//For each outgoing letter...	
	{
	  set<State*> closure = epsilon_closure(target[*it]);
	  State* target_state = get_state(closure,non_det_states);

	  if(target_state != 0)
	    //Target state already exists?
	    result->succ[current_state][new Label(*it)] = target_state;
	      //just add a transition
	  else
	    //Target state does not exists?
	    {
	      State* target_state = new State();
	      non_det_states[target_state] = closure;
	      result->state.push_back(target_state);
	      //Create a new (det) state, and map it to its (non-det) states

	      result->succ[current_state][new Label(*it)] = target_state;
	      //Add a transition labelled by the letter

	      state_stack.push_back(target_state); 
	      //mark target state to be processed
	    }
	}	
    }

  //
  // Find the final states
  //

  State* non_det_final_state = final_state[0];
  //Assumption: only one final state in non-det automaton

  for(map<State*,set<State*> >::const_iterator it = non_det_states.begin(); 
      it != non_det_states.end(); ++it)
    if(it->second.find(non_det_final_state) != it->second.end())
      //non-det final state in the non-det set?
      result->final_state.push_back(it->first);
       //set it as a final state

  return result;
}


void
Automaton::complete()
{
  State* puits = new State();
  state.push_back(puits);

  default_state = puits;

  for(int i=0; i<state.size(); i++)
    //For each state...
    {
      set<char> handled_letters;
      for(map<Label*,State*>::const_iterator it = succ[state[i]].begin(); 
	  it != succ[state[i]].end(); ++it)
	//For each transition...
	if(it->first->is_letter)
	  handled_letters.insert(it->first->letter);
          //...collect handled letters
	else
	  {
	    cerr << "Automaton::completion: epsilon-transitions not allowed"
		 << endl << "Determinize first" << endl;
	    exit(1);
	  }

      for(set<char>::const_iterator it = alphabet.begin(); 
	  it != alphabet.end(); ++it)
	if(handled_letters.find(*it) == handled_letters.end())
	  //For each letter not handled...
	  succ[state[i]][new Label(*it)] = puits;
           //add a transition to the 'puits'
    }

}


Automaton*
Automaton::minimize()
{
  Automaton* result = new Automaton(alphabet);

  //
  // Init distinguishable relation to empty
  //
  vector<vector<int> > distinguishable (state.size());
  for(int i=0; i<state.size(); i++)
    {
      distinguishable[i].resize(state.size());
      for(int j=0; j<state.size(); j++)
	distinguishable[i][j] = false;
    }

  //
  // Mark all pairs (\in F, \notin F) as distinguishable
  //
  for(int i=0; i<state.size(); i++)
    for(int j=0; j<state.size(); j++)
      if((is_final(state[i]) && (!is_final(state[j])))
	 || ((!is_final(state[i])) && is_final(state[j])))
	distinguishable[i][j] = true;

  //back[target_state][letter] = set of previous states
  map<State*,map<char,set<State*> > > back;
  for(int i=0; i<state.size(); i++)
    //For each state...
    {
      map<Label*,State*> target = succ[state[i]];
      for(map<Label*,State*>::const_iterator it = target.begin(); 
	  it != target.end(); ++it)
	//For each out going transition...
	if(it->first->is_letter)
	  back[it->second][it->first->letter].insert(state[i]);
	else
	  {
	    cerr << "Automaton::minimize: epsilon transitions forbidden" << endl
		 << "non-deterministic automaton?" << endl;
	    exit(1);
	  }
    }


  //
  // Core algorithm. 
  // q.a -> q', r.a -> r' /\ (q,r) disting. => (q',r') disting. and so on.
  //
  for(int iter = 0; iter<state.size()*state.size(); iter++)
    for(int i=0; i<state.size(); i++)
      for(int j=0; j<state.size(); j++)
	if(distinguishable[i][j])
	  //(state i, state j) distinguishable?
	  {
	    for(map<char,set<State*> >::const_iterator it = 
		  back[state[i]].begin(); 
		it != back[state[i]].end(); ++it)
	      //For each transition to state i...
	      if(back[state[j]].find(it->first) != back[state[j]].end())
		//do there exist a transition to state j with the same letter?
		{
		  set<State*> back_i_letter = it->second;
		  set<State*> back_j_letter = back[state[j]][it->first];
		  for(set<State*>::const_iterator it1 = back_i_letter.begin(); 
		      it1 != back_i_letter.end(); ++it1)
		    for(set<State*>::const_iterator it2= back_j_letter.begin(); 
			it2 != back_j_letter.end(); ++it2)
		      //For each couple of corresponding backward states...
		      {
			distinguishable[state_rank(*it1)][state_rank(*it2)] =
			  true;
			distinguishable[state_rank(*it2)][state_rank(*it1)] =
			  true;
			//Mark them (in a symmetric manner)
		      }
		}
	  }

  //
  // Minimal automaton construction
  //
  set<set<State*> > eq_class;
  for(int i=0; i<state.size(); i++)
    {
      set<State*> class_i;
      for(int j=0; j<state.size(); j++)
	if(!(distinguishable[i][j]))
	  //For each couple of equivalent states...
	  class_i.insert(state[j]);
      eq_class.insert(class_i);
    }
  //Recover the partition of states

  map<State*,State*> class_state; //"representant"
  for(set<set<State*> >::const_iterator it1=eq_class.begin(); 
      it1 != eq_class.end(); ++it1)
    //For each class...
    {
      State* min_state = new State();
      for(set<State*>::const_iterator it2=it1->begin(); 
	  it2 != it1->end(); ++it2)
	//For each member of the class...
	{
	  class_state[*it2] = min_state;      
	  //cout << state_rank(*it2) << " ";
	}
      //cout << endl;
      result->state.push_back(min_state);
    }
  //Map each state of a partition to a unique state


  map<State*,set<char> > added_letters; //to avoid redondant transitions

  for(int i=0; i<state.size(); i++)
    //For each state...
    {
      set<char> existing_letter = added_letters[class_state[state[i]]]; 
      //added from class(node_i)
      
      for(map<Label*,State*>::const_iterator it = succ[state[i]].begin(); 
	  it != succ[state[i]].end();  ++it)
	//For each transition starting from that state...
	if(it->first->is_letter)
	  {
	    if(existing_letter.find(it->first->letter) == existing_letter.end())
	      //Letter not yet handled?
	      result->succ[class_state[state[i]]][it->first] = 
		class_state[it->second]; //Add the transition
	    
	    added_letters[class_state[state[i]]].insert(it->first->letter);
	    //let the letter exists from class(node_i)
	  }
	else
	  {
	    cerr << "Automaton::minimize: epsilon-transition not handled"
		 << "Determinize first" << endl;
	    exit(1);
	  }
    }
  //Add the quotient of the transition relation.

  result->initial_state = class_state[initial_state];
  //initial state = class(initial_state)

  for(int i=0; i<final_state.size(); i++)
    result->final_state.push_back(class_state[final_state[i]]);
  //final states = class(final states)

  result->default_state = class_state[default_state];
  //default state = class(default state)

  return result;
}


pair<pair< vector<vector<int> > , map<char,int> > , vector<int> >
Automaton::transition_table(set<char> markers)
{
  vector<vector<int> > result ( state.size() );

  map<char,int> letter_rank;
  int rank = 0;
  for(set<char>::const_iterator it = alphabet.begin(); 
      it != alphabet.end(); ++it)
    if(markers.find(*it) == markers.end())
      letter_rank[*it] = rank++;
  //Assign a rank to each (non-marker) letter

  vector<int> state_marker (state.size());
  for(int i=0; i<state.size(); i++)
    state_marker[i] = 0; //error code
  //Last marker for each state

  for(int i=0; i<state.size(); i++)
    //For each state...
    {
      result[i].resize(rank);
      for(map<Label*,State*>::const_iterator it = succ[state[i]].begin(); 
	  it != succ[state[i]].end(); ++it)
	//For each transition starting from that state...
	if(it->first->is_letter) 
	  {
	    char letter = it->first->letter;
	    if(markers.find(letter) == markers.end())
	      //Not a marker?
	      result[i][letter_rank[letter]] = state_rank(it->second);
	    else 
	      //Marker => the smallest one is prioritary...
	      {
		if(it->second == initial_state)
		  if(state_marker[i]>(int)letter || state_marker[i] == 0)
		    state_marker[i] = (int)letter;
	      }
	  }
	else
	  {
	    cerr << "Automaton::transition_table: epsilon-transition not "
		 << "handled" << endl << "Determinize first" << endl;
	    exit(1);
	  }
    }

  pair<pair< vector<vector<int> > , map<char,int> > , vector<int> >
    final_result;
  final_result.first.first = result; //transition table itself
  final_result.first.second = letter_rank; //mapping letter -> column
  final_result.second = state_marker; //mapping state -> token 

  return final_result;
}


void 
Automaton::print_dot(ostream& sout)
{
  sout << "digraph g {" << endl;

  //Nodes
  for(int i=0; i<state.size(); i++)
    //For each state...
    {
      sout << "  state" << i << "[shape=box, label=\"" << i;
      if(initial_state == state[i])
	sout << "\\nstart";
      for(int j=0; j<final_state.size(); j++)
	if(final_state[j] == state[i])
	  sout << "\\nstop";
      sout << "\"];" << endl;
    }

  //Transitions
  for(int i=0; i<state.size(); i++)
    //For each state...
    {
      for(map<Label*,State*>::const_iterator it = succ[state[i]].begin(); 
	  it != succ[state[i]].end(); ++it)
	//For each transition starting from 'state'...
	sout << "  state" << i << " -> state" << state_rank(it->second)
	     << "[label=\"" << *(it->first) << "\"];" << endl;
    }

  sout << "}" << endl;
}


ostream& operator <<(ostream& sout, Label& label)
{
  if(label.is_epsilon)
    sout << "epsilon";

  if(label.is_letter)
    {
      if(label.letter == ' ')
	sout << "' '";
      else
	{
	  switch(label.letter) {
	  case '\0': { sout << "-0-"; break; }
	  case '\1': { sout << "-1-"; break; }
	  case '\2': { sout << "-2-"; break; }
	  case '\3': { sout << "-3-"; break; }
	  case '\4': { sout << "-4-"; break; }
	  case '\5': { sout << "-5-"; break; }
	  case '\6': { sout << "-6-"; break; }
	  case '\7': { sout << "-7-"; break; }
	  default: { sout << label.letter; break; }
	  }
	}
    }

  return sout;
}
